# -*- coding: utf-8 -*-
#
# Copyright 2017-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Storage command tests."""

import os
import subprocess
from pathlib import Path

from renku.cli import cli


def test_lfs_storage_clean_no_remote(runner, project, client):
    """Test ``renku storage clean`` command with no remote set."""
    with (client.path / 'tracked').open('w') as fp:
        fp.write('tracked file')
    client.repo.git.add('*')
    client.repo.index.commit('tracked file')

    result = runner.invoke(
        cli, ['storage', 'clean', 'tracked'], catch_exceptions=False
    )
    assert 1 == result.exit_code
    assert 'No git remote is configured for' in result.output


def test_lfs_storage_clean(runner, project, client_with_remote):
    """Test ``renku storage clean`` command."""
    client = client_with_remote['client']

    with (client.path / 'tracked').open('w') as fp:
        fp.write('tracked file')
    client.repo.git.add('*')
    client.repo.index.commit('tracked file')

    result = runner.invoke(
        cli, ['storage', 'clean', 'tracked'], catch_exceptions=False
    )
    assert 0 == result.exit_code
    assert 'These paths were ignored as they are not tracked' in result.output

    subprocess.call(['git', 'lfs', 'track', 'tracked'])
    client.repo.git.add('*')
    client.repo.index.commit('Tracked in lfs')
    client.repo.git.push('--no-verify', 'origin')

    with (client.path / 'tracked').open('r') as fp:
        assert 'tracked file' in fp.read()

    assert 'tracked' in Path('.gitattributes').read_text()

    lfs_objects = []
    for _, _, files in os.walk(str(client.path / '.git' / 'lfs' / 'objects')):
        lfs_objects.extend(files)

    assert 1 == len(lfs_objects)

    result = runner.invoke(
        cli, ['storage', 'clean', 'tracked'], catch_exceptions=False
    )
    assert 0 == result.exit_code

    assert 'version https://git-lfs.github.com/spec/v1' in (
        client.path / 'tracked'
    ).read_text()

    lfs_objects = []
    for _, _, files in os.walk(str(client.path / '.git' / 'lfs' / 'objects')):
        lfs_objects.extend(files)

    assert 0 == len(lfs_objects)

    # already clean file should be ignored on clean
    result = runner.invoke(
        cli, ['storage', 'clean', 'tracked'], catch_exceptions=False
    )
    assert 0 == result.exit_code


def test_lfs_storage_unpushed_clean(runner, project, client_with_remote):
    """Test ``renku storage clean`` command for unpushed files."""
    client = client_with_remote['client']

    with (client.path / 'tracked').open('w') as fp:
        fp.write('tracked file')
    subprocess.call(['git', 'lfs', 'track', 'tracked'])
    client.repo.git.add('*')
    client.repo.index.commit('tracked file')

    result = runner.invoke(
        cli, ['storage', 'clean', 'tracked'], catch_exceptions=False
    )

    assert 0 == result.exit_code
    assert 'These paths were ignored as they are not pushed' in result.output



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Dataverse API integration."""
import json
import pathlib
import re
import urllib
import urllib.parse as urlparse
from pathlib import Path
from string import Template

import attr
import requests
from tqdm import tqdm

from renku.core import errors
from renku.core.commands.providers.api import ExporterApi, ProviderApi
from renku.core.commands.providers.doi import DOIProvider
from renku.core.models.datasets import Dataset, DatasetFile
from renku.core.utils.doi import extract_doi, is_doi
from renku.core.utils.requests import retry

from .dataverse_metadata_templates import AUTHOR_METADATA_TEMPLATE, \
    CONTACT_METADATA_TEMPLATE, DATASET_METADATA_TEMPLATE

DATAVERSE_API_PATH = 'api/v1'

DATAVERSE_VERSION_API = 'info/version'
DATAVERSE_METADATA_API = 'datasets/export'
DATAVERSE_FILE_API = 'access/datafile/:persistentId/'
DATAVERSE_EXPORTER = 'schema.org'


def check_dataverse_uri(url):
    """Check if an URL points to a dataverse instance."""
    url_parts = list(urlparse.urlparse(url))
    url_parts[2] = pathlib.posixpath.join(
        DATAVERSE_API_PATH, DATAVERSE_VERSION_API
    )

    url_parts[3:6] = [''] * 3
    version_url = urlparse.urlunparse(url_parts)

    with retry() as session:
        response = session.get(version_url)

        if response.status_code != 200:
            return False

        version_data = response.json()

        if 'status' not in version_data or 'data' not in version_data:
            return False

        version_info = version_data['data']

        if 'version' not in version_info or 'build' not in version_info:
            return False

        return True


def check_dataverse_doi(doi):
    """Check if a DOI points to a dataverse dataset."""
    try:
        doi = DOIProvider().find_record(doi)
    except LookupError:
        return False

    return check_dataverse_uri(doi.url)


def make_records_url(record_id, base_url):
    """Create URL to access record by ID."""
    url_parts = list(urlparse.urlparse(base_url))
    url_parts[2] = pathlib.posixpath.join(
        DATAVERSE_API_PATH, DATAVERSE_METADATA_API
    )
    args_dict = {'exporter': DATAVERSE_EXPORTER, 'persistentId': record_id}
    url_parts[4] = urllib.parse.urlencode(args_dict)
    return urllib.parse.urlunparse(url_parts)


def make_file_url(file_id, base_url):
    """Create URL to access record by ID."""
    url_parts = list(urlparse.urlparse(base_url))
    url_parts[2] = pathlib.posixpath.join(
        DATAVERSE_API_PATH, DATAVERSE_FILE_API
    )
    args_dict = {'persistentId': file_id}
    url_parts[4] = urllib.parse.urlencode(args_dict)
    return urllib.parse.urlunparse(url_parts)


@attr.s
class DataverseProvider(ProviderApi):
    """Dataverse API provider."""

    is_doi = attr.ib(default=False)

    _accept = attr.ib(default='application/json')

    _server_url = attr.ib(default=None)

    _dataverse_name = attr.ib(default=None)

    @staticmethod
    def supports(uri):
        """Check if provider supports a given uri."""
        is_doi_ = is_doi(uri)

        is_dataverse_uri = is_doi_ is None and check_dataverse_uri(uri)
        is_dataverse_doi = is_doi_ and check_dataverse_doi(is_doi_.group(0))

        return is_dataverse_uri or is_dataverse_doi

    @staticmethod
    def record_id(uri):
        """Extract record id from uri."""
        parsed = urlparse.urlparse(uri)
        return urlparse.parse_qs(parsed.query)['persistentId'][0]

    def _make_request(self, uri):
        """Execute network request."""
        with retry() as session:
            response = session.get(uri, headers={'Accept': self._accept})
            if response.status_code != 200:
                raise LookupError(
                    'record not found. Status: {}'.format(
                        response.status_code
                    )
                )
            return response

    def find_record(self, uri, client=None):
        """Retrieves a record from Dataverse.

        :raises: ``LookupError``
        :param uri: DOI or URL
        :return: ``DataverseRecord``
        """
        if self.is_doi:
            doi = DOIProvider().find_record(uri)
            uri = doi.url

        uri = self._get_export_uri(uri)

        return self._get_record(uri)

    def _get_export_uri(self, uri):
        """Gets a dataverse api export URI from a dataverse entry."""
        record_id = DataverseProvider.record_id(uri)
        uri = make_records_url(record_id, uri)
        return uri

    def _get_record(self, uri):
        """Retrieve metadata and return ``DataverseRecordSerializer``."""
        response = self._make_request(uri)

        return DataverseRecordSerializer(
            json=response.json(), dataverse=self, uri=uri
        )

    def get_exporter(self, dataset, access_token):
        """Create export manager for given dataset."""
        return DataverseExporter(
            dataset=dataset,
            access_token=access_token,
            server_url=self._server_url,
            dataverse_name=self._dataverse_name
        )

    def set_parameters(
        self, client, *, dataverse_server_url, dataverse_name, **kwargs
    ):
        """Set and validate required parameters for a provider."""
        CONFIG_BASE_URL = 'server_url'

        if not dataverse_server_url:
            dataverse_server_url = client.get_value(
                'dataverse', CONFIG_BASE_URL
            )
        else:
            client.set_value(
                'dataverse',
                CONFIG_BASE_URL,
                dataverse_server_url,
                global_only=True
            )

        if not dataverse_server_url:
            raise errors.ParameterError('Dataverse server URL is required.')

        if not dataverse_name:
            raise errors.ParameterError('Dataverse name is required.')

        self._server_url = dataverse_server_url
        self._dataverse_name = dataverse_name


@attr.s
class DataverseRecordSerializer:
    """Dataverse record Serializer."""

    _dataverse = attr.ib(default=None, kw_only=True)

    _uri = attr.ib(default=None, kw_only=True)

    _json = attr.ib(default=None, kw_only=True)

    def is_last_version(self, uri):
        """Check if record is at last possible version."""
        return True

    def _convert_json_property_name(self, property_name):
        """Removes '@' and converts names to snake_case."""
        property_name = property_name.strip('@')
        property_name = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', property_name)
        return re.sub('([a-z0-9])([A-Z])', r'\1_\2', property_name).lower()

    @property
    def files(self):
        """Get all file metadata entries."""
        file_list = []

        for f in self._json['distribution']:
            mapped_file = {
                self._convert_json_property_name(k): v
                for k, v in f.items()
            }
            mapped_file['parent_url'] = self._uri
            file_list.append(mapped_file)
        return file_list

    def get_files(self):
        """Get Dataverse files metadata as ``DataverseFileSerializer``."""
        if not self.files:
            raise LookupError(
                'no files have been found - deposit is empty or protected'
            )

        return [DataverseFileSerializer(**file_) for file_ in self.files]

    def as_dataset(self, client):
        """Deserialize `DataverseRecordSerializer` to `Dataset`."""
        files = self.get_files()
        dataset = Dataset.from_jsonld(self._json, client=client)

        if dataset.description and not dataset.description.strip():
            dataset.description = None

        for creator in dataset.creator:
            if creator.affiliation == '':
                creator.affiliation = None

        serialized_files = []
        for file_ in files:
            remote_ = file_.remote_url
            dataset_file = DatasetFile(
                url=remote_.geturl(),
                id=file_._id if file_._id else file_.name,
                filename=file_.name,
                filesize=file_.content_size,
                filetype=file_.file_format,
                path='',
            )
            serialized_files.append(dataset_file)

        dataset.files = serialized_files

        return dataset


@attr.s
class DataverseFileSerializer:
    """Dataverse record file."""

    _id = attr.ib(default=None, kw_only=True)

    identifier = attr.ib(default=None, kw_only=True)

    name = attr.ib(default=None, kw_only=True)

    file_format = attr.ib(default=None, kw_only=True)

    content_size = attr.ib(default=None, kw_only=True)

    description = attr.ib(default=None, kw_only=True)

    content_url = attr.ib(default=None, kw_only=True)

    parent_url = attr.ib(default=None, kw_only=True)

    _type = attr.ib(default=None, kw_only=True)

    @property
    def remote_url(self):
        """Get remote URL as ``urllib.ParseResult``."""
        if self.content_url is not None:
            return urllib.parse.urlparse(self.content_url)

        if self.identifier is None:
            return None

        doi = extract_doi(self.identifier)

        if doi is None:
            return None

        file_url = make_file_url('doi:' + doi, self.parent_url)

        return urllib.parse.urlparse(file_url)


@attr.s
class DataverseExporter(ExporterApi):
    """Dataverse export manager."""

    dataset = attr.ib(kw_only=True)

    access_token = attr.ib(kw_only=True)

    _server_url = attr.ib(kw_only=True, default=None)

    _dataverse_name = attr.ib(kw_only=True, default=None)

    def set_access_token(self, access_token):
        """Set access token."""
        self.access_token = access_token

    def access_token_url(self):
        """Endpoint for creation of access token."""
        return urllib.parse.urljoin(
            self._server_url, '/dataverseuser.xhtml?selectTab=apiTokenTab'
        )

    def export(self, publish, **kwargs):
        """Execute export process."""
        deposition = _DataverseDeposition(
            server_url=self._server_url, access_token=self.access_token
        )
        metadata = self._get_dataset_metadata()
        response = deposition.create_dataset(
            dataverse_name=self._dataverse_name, metadata=metadata
        )
        dataset_pid = response.json()['data']['persistentId']

        with tqdm(total=len(self.dataset.files)) as progressbar:
            for file_ in self.dataset.files:
                try:
                    path = Path(file_.path).relative_to(self.dataset.datadir)
                except ValueError:
                    path = Path(file_.path)
                deposition.upload_file(
                    full_path=file_.full_path, path_in_dataset=path
                )
                progressbar.update(1)

        if publish:
            deposition.publish_dataset()

        return dataset_pid

    def _get_dataset_metadata(self):
        authors, contacts = self._get_creators()
        metadata_template = Template(DATASET_METADATA_TEMPLATE)
        metadata = metadata_template.substitute(
            name=_escape_json_string(self.dataset.name),
            authors=json.dumps(authors),
            contacts=json.dumps(contacts),
            description=_escape_json_string(self.dataset.description)
        )
        return json.loads(metadata)

    def _get_creators(self):
        authors = []
        contacts = []

        for creator in self.dataset.creator:
            name = creator.name or ''
            affiliation = creator.affiliation or ''
            email = creator.email or ''

            author_template = Template(AUTHOR_METADATA_TEMPLATE)
            author = author_template.substitute(
                name=_escape_json_string(name),
                affiliation=_escape_json_string(affiliation)
            )
            authors.append(json.loads(author))

            contact_template = Template(CONTACT_METADATA_TEMPLATE)
            contact = contact_template.substitute(
                name=_escape_json_string(name), email=email
            )
            contacts.append(json.loads(contact))

        return authors, contacts


@attr.s
class _DataverseDeposition:
    """Dataverse record for deposit."""

    access_token = attr.ib(kw_only=True)
    server_url = attr.ib(kw_only=True)
    dataset_pid = attr.ib(kw_only=True, default=None)

    DATASET_CREATE_PATH = 'dataverses/{dataverseName}/datasets'
    FILE_UPLOAD_PATH = 'datasets/:persistentId/add'
    DATASET_PUBLISH_PATH = 'datasets/:persistentId/actions/:publish'

    def create_dataset(self, dataverse_name, metadata):
        """Create a dataset in a given dataverse."""
        api_path = self.DATASET_CREATE_PATH.format(
            dataverseName=dataverse_name
        )
        url = self._make_url(api_path=api_path)

        response = self._post(url=url, json=metadata)
        self._check_response(response)

        self.dataset_pid = response.json()['data']['persistentId']

        return response

    def upload_file(self, full_path, path_in_dataset):
        """Upload a file to a previously-created dataset."""
        if self.dataset_pid is None:
            raise errors.ExportError('Dataset not created.')

        url = self._make_url(
            self.FILE_UPLOAD_PATH, persistentId=self.dataset_pid
        )

        params = {'directoryLabel': str(path_in_dataset.parent)}
        data = dict(jsonData=json.dumps(params))

        files = {'file': (path_in_dataset.name, open(full_path, 'rb'))}

        response = self._post(url=url, data=data, files=files)
        self._check_response(response)

        return response

    def publish_dataset(self):
        """Publish a previously-created dataset."""
        if self.dataset_pid is None:
            raise errors.ExportError('Dataset not created.')

        url = self._make_url(
            self.DATASET_PUBLISH_PATH,
            persistentId=self.dataset_pid,
            type='major'
        )

        response = self._post(url=url)
        self._check_response(response)

        return response

    def _make_url(self, api_path, **query_params):
        """Create URL for creating a dataset."""
        url_parts = urlparse.urlparse(self.server_url)
        path = pathlib.posixpath.join(DATAVERSE_API_PATH, api_path)

        query_params = urllib.parse.urlencode(query_params)
        url_parts = url_parts._replace(path=path, query=query_params)
        return urllib.parse.urlunparse(url_parts)

    def _post(self, url, json=None, data=None, files=None):
        headers = {'X-Dataverse-key': self.access_token}
        try:
            with retry() as session:
                return session.post(
                    url=url,
                    json=json,
                    data=data,
                    files=files,
                    headers=headers
                )
        except requests.exceptions.RequestException as e:
            raise errors.ExportError('Cannot POST to remote server.') from e

    @staticmethod
    def _check_response(response):
        if response.status_code not in [200, 201]:
            if response.status_code == 401:
                raise errors.AuthenticationError(
                    'Access unauthorized - update access token.'
                )

            raise errors.ExportError(
                'HTTP {} - Cannot export dataset: {}'.format(
                    response.status_code,
                    response.json()['message']
                )
            )


def _escape_json_string(value):
    """Create a JSON-safe string."""
    if isinstance(value, str):
        return json.dumps(value)[1:-1]
    return value



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service cache management for files."""
from renku.service.cache.files import FileManagementCache
from renku.service.cache.jobs import JobManagementCache
from renku.service.cache.projects import ProjectManagementCache
from renku.service.cache.users import UserManagementCache
from renku.service.config import CACHE_PROJECTS_PATH, CACHE_UPLOADS_PATH


class ServiceCache(
    FileManagementCache, ProjectManagementCache, JobManagementCache,
    UserManagementCache
):
    """Service cache manager."""

    pass


def make_cache():
    """Create cache structure."""
    sub_dirs = [CACHE_UPLOADS_PATH, CACHE_PROJECTS_PATH]

    for subdir in sub_dirs:
        subdir.mkdir(parents=True, exist_ok=True)

    return ServiceCache()


cache = make_cache()

__all__ = ['cache', 'ServiceCache']



# -*- coding: utf-8 -*-
#
# Copyright 2017-%d - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Version information for Renku."""

__version__ = '0.10.4.dev31'


def _get_disribution_url():
    try:
        import pkg_resources
        d = pkg_resources.get_distribution('renku')
        metadata = d._get_metadata(d.PKG_INFO)
        home_page = [m for m in metadata if m.startswith('Home-page:')]
        return home_page[0].split(':', 1)[1].strip()
    except Exception:
        return 'N/A'


version_url = '{}/tree/{}'.format(_get_disribution_url(), 'v' + __version__)



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service utility functions."""
import uuid

from git import GitCommandError, Repo

from renku.service.config import CACHE_PROJECTS_PATH, CACHE_UPLOADS_PATH


def make_project_path(user, project):
    """Construct full path for cached project."""
    valid_user = user and 'user_id' in user
    valid_project = project and 'owner' in project and 'name' in project

    if valid_user and valid_project:
        return (
            CACHE_PROJECTS_PATH / user['user_id'] / project['owner'] /
            project['name']
        )


def make_file_path(user, cached_file):
    """Construct full path for cache file."""
    valid_user = user and 'user_id' in user
    valid_file = cached_file and 'file_name' in cached_file

    if valid_user and valid_file:
        return (
            CACHE_UPLOADS_PATH / user['user_id'] / cached_file['relative_path']
        )


def valid_file(user, cached_file):
    """Ensure file system and cache state matches."""
    file_path = make_file_path(user, cached_file)

    if file_path.exists():
        cached_file['is_dir'] = file_path.is_dir()
        return cached_file


def repo_sync(repo_path, remote_names=('origin', )):
    """Sync the repo with the remotes."""
    repo = Repo(repo_path)
    pushed_branch = None

    for remote in repo.remotes:
        if remote.name in remote_names:
            try:
                repo.git.push(remote.name, repo.active_branch)
                pushed_branch = repo.active_branch
            except GitCommandError as e:
                if 'protected branches' not in e.stderr:
                    raise e

                pushed_branch = uuid.uuid4().hex
                repo.git.checkout(b=pushed_branch)
                repo.git.push(remote.name, repo.active_branch)

    return pushed_branch



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku migrations management.

Migrations files are put in renku/core/management/migrations directory. Name
of these files has m_1234__name.py format where 1234 is the migration version
and name can be any alphanumeric and underscore combination. Migration files
are sorted based on their lowercase name. Each migration file must define a
public "migrate" function that accepts a client as its argument.

When executing a migration, the migration file is imported as a module and the
"migrate" function is executed. Migration version is checked against the Renku
project version (in .renku/metadata.yml) and any migration which has a higher
version is applied to the project.
"""
import importlib
import re
from pathlib import Path

import pkg_resources

SUPPORTED_PROJECT_VERSION = 4


def is_migration_required(client):
    """Check if project requires migration."""
    return (
        _is_renku_project(client) and
        _get_project_version(client) < SUPPORTED_PROJECT_VERSION
    )


def is_project_unsupported(client):
    """Check if this version of Renku cannot work with the project."""
    return (
        _is_renku_project(client) and
        _get_project_version(client) > SUPPORTED_PROJECT_VERSION
    )


def migrate(client, progress_callback=None):
    """Apply all migration files to the project."""
    if not _is_renku_project(client):
        return

    project_version = _get_project_version(client)
    n_migrations_executed = 0

    for version, path in get_migrations():
        if version > project_version:
            module = importlib.import_module(path)
            if progress_callback:
                module_name = module.__name__.split('.')[-1]
                progress_callback(f'Applying migration {module_name}...')
            module.migrate(client)
            n_migrations_executed += 1

    client.project.version = str(version)
    client.project.to_yaml()

    if progress_callback and n_migrations_executed > 0:
        progress_callback(
            f'Successfully applied {n_migrations_executed} migrations.'
        )

    return n_migrations_executed != 0


def _get_project_version(client):
    try:
        return int(client.project.version)
    except ValueError:
        return 1


def _is_renku_project(client):
    return client.project is not None


def get_migrations():
    """Return a sorted list of versions and migration modules."""
    migrations = []
    for file_ in pkg_resources.resource_listdir(
        'renku.core.management', 'migrations'
    ):
        match = re.search(r'm_([0-9]{4})__[a-zA-Z0-9_-]*.py', file_)

        if match is None:  # migration files match m_0000__[name].py format
            continue

        version = int(match.groups()[0])
        path = 'renku.core.management.migrations.{}'.format(Path(file_).stem)
        migrations.append((version, path))

    migrations = sorted(migrations, key=lambda v: v[1].lower())
    return migrations



# -*- coding: utf-8 -*-
#
# Copyright 2018-2020- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Serializers for graph data."""

import functools

import click

from renku.core.errors import SHACLValidationError
from renku.core.utils.shacl import validate_graph


def ascii(graph, strict=False):
    """Format graph as an ASCII art."""
    from ..ascii import DAG
    from ..echo import echo_via_pager

    if strict:
        raise SHACLValidationError('--strict not supported for json-ld-graph')

    echo_via_pager(str(DAG(graph)))


def _jsonld(graph, format, *args, **kwargs):
    """Return formatted graph in JSON-LD ``format`` function."""
    import json

    from renku.core.compat import pyld
    from renku.core.models.jsonld import asjsonld

    output = getattr(pyld.jsonld, format)([
        asjsonld(action) for action in graph.activities.values()
    ])
    return json.dumps(output, indent=2)


def _conjunctive_graph(graph):
    """Convert a renku ``Graph`` to an rdflib ``ConjunctiveGraph``."""
    from rdflib import ConjunctiveGraph
    from rdflib.plugin import register, Parser

    register('json-ld', Parser, 'rdflib_jsonld.parser', 'JsonLDParser')

    return ConjunctiveGraph().parse(
        data=_jsonld(graph, 'expand'),
        format='json-ld',
    )


def dot(graph, simple=True, debug=False, landscape=False, strict=False):
    """Format graph as a dot file."""
    import sys

    from rdflib.tools.rdf2dot import rdf2dot

    if strict:
        raise SHACLValidationError('--strict not supported for dot')

    g = _conjunctive_graph(graph)

    g.bind('prov', 'http://www.w3.org/ns/prov#')
    g.bind('foaf', 'http://xmlns.com/foaf/0.1/')
    g.bind('wfdesc', 'http://purl.org/wf4ever/wfdesc#')
    g.bind('wf', 'http://www.w3.org/2005/01/wf/flow#')
    g.bind('wfprov', 'http://purl.org/wf4ever/wfprov#')
    g.bind('schema', 'http://schema.org/')

    if debug:
        rdf2dot(g, sys.stdout)
        return

    sys.stdout.write('digraph { \n node [ fontname="DejaVu Sans" ] ; \n ')
    if landscape:
        sys.stdout.write('rankdir="LR" \n')
    if simple:
        _rdf2dot_simple(g, sys.stdout)
        return
    _rdf2dot_reduced(g, sys.stdout)


# define the various dot options
dot_full = functools.partial(dot, simple=False, landscape=False)
dot_landscape = functools.partial(dot, simple=True, landscape=True)
dot_full_landscape = functools.partial(dot, simple=False, landscape=True)
dot_debug = functools.partial(dot, debug=True)


def _rdf2dot_simple(g, stream):
    """Create a simple graph of processes and artifacts."""
    from itertools import chain

    import re

    path_re = re.compile(
        r'(?P<prefix>https://\w+/|https://\w+/\w+/){0,1}(?P<type>[a-zA-Z]+)/'
        r'(?P<commit>\w+)'
        r'(?P<path>.+)?'
    )

    inputs = g.query(
        """
        SELECT ?input ?role ?activity ?comment
        WHERE {
            ?activity (prov:qualifiedUsage/prov:entity) ?input .
            ?activity prov:qualifiedUsage ?qual .
            ?qual prov:hadRole ?role .
            ?qual prov:entity ?input .
            ?qual rdf:type ?type .
            ?activity rdf:type wfprov:ProcessRun .
            ?activity rdfs:comment ?comment .
            FILTER NOT EXISTS {?activity rdf:type wfprov:WorkflowRun}
        }
        """
    )
    outputs = g.query(
        """
        SELECT ?activity ?role ?output ?comment
        WHERE {
            ?output (prov:qualifiedGeneration/prov:activity) ?activity .
            ?output prov:qualifiedGeneration ?qual .
            ?qual prov:hadRole ?role .
            ?qual prov:activity ?activity .
            ?qual rdf:type ?type .
            ?activity rdf:type wfprov:ProcessRun ;
                      rdfs:comment ?comment .
            FILTER NOT EXISTS {?activity rdf:type wfprov:WorkflowRun}
        }
        """
    )

    activity_nodes = {}
    artifact_nodes = {}
    for source, role, target, comment, in chain(inputs, outputs):
        # extract the pieces of the process URI
        src_path = path_re.match(source).groupdict()
        tgt_path = path_re.match(target).groupdict()

        # write the edge
        stream.write(
            '\t"{src_commit}:{src_path}" -> '
            '"{tgt_commit}:{tgt_path}" '
            '[label={role}] \n'.format(
                src_commit=src_path['commit'][:5],
                src_path=src_path.get('path') or '',
                tgt_commit=tgt_path['commit'][:5],
                tgt_path=tgt_path.get('path') or '',
                role=role
            )
        )
        if src_path.get('type') == 'commit':
            activity_nodes.setdefault(source, {'comment': comment})
            artifact_nodes.setdefault(target, {})
        if tgt_path.get('type') == 'commit':
            activity_nodes.setdefault(target, {'comment': comment})
            artifact_nodes.setdefault(source, {})

    # customize the nodes
    for node, content in activity_nodes.items():
        node_path = path_re.match(node).groupdict()
        stream.write(
            '\t"{commit}:{path}" '
            '[shape=box label="#{commit}:{path}:{comment}"] \n'.format(
                comment=content['comment'],
                commit=node_path['commit'][:5],
                path=node_path.get('path') or ''
            )
        )
    for node, content in artifact_nodes.items():
        node_path = path_re.match(node).groupdict()
        stream.write(
            '\t"{commit}:{path}" '
            '[label="#{commit}:{path}"] \n'.format(
                commit=node_path['commit'][:5],
                path=node_path.get('path') or ''
            )
        )
    stream.write('}\n')


def _rdf2dot_reduced(g, stream):
    """
    A reduced dot graph.

    Adapted from original source:
    https://rdflib.readthedocs.io/en/stable/_modules/rdflib/tools/rdf2dot.html
    """
    import html
    import collections
    import rdflib

    from rdflib.tools.rdf2dot import LABEL_PROPERTIES, NODECOLOR

    types = collections.defaultdict(set)
    fields = collections.defaultdict(set)
    nodes = {}

    def node(x):
        """Return a name of the given node."""
        return nodes.setdefault(x, 'node{0}'.format(len(nodes)))

    def label(x, g):
        """Generate a label for the node."""
        for labelProp in LABEL_PROPERTIES:
            label_ = g.value(x, labelProp)
            if label_:
                return label_

        try:
            return g.namespace_manager.compute_qname(x)[2]
        except Exception:
            return x

    def formatliteral(l, g):
        """Format and escape literal."""
        v = html.escape(l)
        if l.datatype:
            return '&quot;%s&quot;^^%s' % (v, qname(l.datatype, g))
        elif l.language:
            return '&quot;%s&quot;@%s' % (v, l.language)
        return '&quot;%s&quot;' % v

    def qname(x, g):
        """Compute qname."""
        try:
            q = g.compute_qname(x)
            return q[0] + ':' + q[2]
        except Exception:
            return x

    def color(p):
        """Choose node color."""
        return 'BLACK'

    # filter out nodes and edges created for directories
    sparql = """
    SELECT ?s ?p ?o
    WHERE {
        ?s ?p ?o
        MINUS {
            ?s rdf:type prov:Collection.
        }
        MINUS {
            VALUES ?exclude { prov:wasInformedBy prov:influenced rdf:label }
            ?s ?exclude ?o.
        }
    }
    """

    for s, p, o in g.query(sparql):
        sn = node(s)
        if p == rdflib.RDFS.label:
            continue
        # inject the type predicate into the node itself
        if p == rdflib.RDF.type:
            types[sn].add((qname(p, g), html.escape(o)))
            continue
        # add the project membership to the node
        if p == rdflib.term.URIRef('schema:isPartOf'):
            fields[sn].add((qname(p, g), html.escape(o)))
            continue

        if isinstance(o, (rdflib.URIRef, rdflib.BNode)):
            on = node(o)
            opstr = (
                '\t%s -> %s [ color=%s, label=< <font point-size="12" '
                'color="#336633">%s</font> > ] ;\n'
            )
            stream.write(opstr % (sn, on, color(p), qname(p, g)))
        else:
            fields[sn].add((qname(p, g), formatliteral(o, g)))

    for u, n in nodes.items():
        stream.write(u'# %s %s\n' % (u, n))
        f = [
            '<tr><td align="left"><b>%s</b></td><td align="left">'
            '<b>%s</b></td></tr>' % x for x in sorted(types[n])
        ]
        f += [
            '<tr><td align="left">%s</td><td align="left">%s</td></tr>' % x
            for x in sorted(fields[n])
        ]
        opstr = (
            '%s [ shape=none, color=%s label=< <table color="#666666"'
            ' cellborder="0" cellspacing="0" border="1"><tr>'
            '<td colspan="2" bgcolor="grey"><B>%s</B></td></tr><tr>'
            '<td href="%s" bgcolor="#eeeeee" colspan="2">'
            '<font point-size="12" color="#6666ff">%s</font></td>'
            '</tr>%s</table> > ] \n'
        )
        stream.write(opstr % (n, NODECOLOR, label(u, g), u, u, ''.join(f)))

    stream.write('}\n')


def makefile(graph, strict=False):
    """Format graph as Makefile."""
    from renku.core.models.provenance.activities import ProcessRun, WorkflowRun

    if strict:
        raise SHACLValidationError('--strict not supported for json-ld-graph')

    for activity in graph.activities.values():
        if not isinstance(activity, ProcessRun):
            continue
        elif isinstance(activity, WorkflowRun):
            steps = activity.subprocesses.values()
        else:
            steps = [activity]

        for step in steps:
            click.echo(' '.join(step.outputs) + ': ' + ' '.join(step.inputs))
            tool = step.process
            click.echo(
                '\t@' + ' '.join(tool.to_argv()) + ' ' + ' '.join(
                    tool.STD_STREAMS_REPR[key] + ' ' + str(path)
                    for key, path in tool._std_streams().items()
                )
            )


def jsonld(graph, strict=False):
    """Format graph as JSON-LD file."""
    ld = _jsonld(graph, 'expand')

    if strict:
        r, _, t = validate_graph(ld, format='json-ld')

        if not r:
            raise SHACLValidationError(
                "{}\nCouldn't get log: Invalid Knowledge Graph data".format(t)
            )
    click.echo(ld)


def jsonld_graph(graph, strict=False):
    """Format graph as JSON-LD graph file."""
    if strict:
        raise SHACLValidationError('--strict not supported for json-ld-graph')
    click.echo(_jsonld(graph, 'flatten'))


def nt(graph, strict=False):
    """Format graph as n-tuples."""
    nt = _conjunctive_graph(graph).serialize(format='nt')
    if strict:
        r, _, t = validate_graph(nt, format='nt')

        if not r:
            raise SHACLValidationError(
                "{}\nCouldn't get log: Invalid Knowledge Graph data".format(t)
            )

    click.echo(nt)


def rdf(graph, strict=False):
    """Output the graph as RDF."""
    xml = _conjunctive_graph(graph).serialize(format='application/rdf+xml')
    if strict:
        r, _, t = validate_graph(xml, format='xml')

        if not r:
            raise SHACLValidationError(
                "{}\nCouldn't get log: Invalid Knowledge Graph data".format(t)
            )

    click.echo(xml)


FORMATS = {
    'ascii': ascii,
    'dot': dot,
    'dot-full': dot_full,
    'dot-landscape': dot_landscape,
    'dot-full-landscape': dot_full_landscape,
    'dot-debug': dot_debug,
    'json-ld': jsonld,
    'json-ld-graph': jsonld_graph,
    'Makefile': makefile,
    'nt': nt,
    'rdf': rdf,
}
"""Valid formatting options."""
